# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

import datetime
from enum import auto
from enum import Enum

import googlemaps

from leistungsbot import leistungs_config as lc


class Openness(Enum):
    UNKNOWN = auto()
    CLOSED = auto()
    SHORT = auto()
    OPEN = auto()


class Weekday(Enum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THIRSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    def toIsoWeekday(self):
        return self.value + 1

    def toPythonWeekday(self):
        return self.value

    def toGooglePlacesWeekday(self):
        return (self.value + 1) % 7

    @staticmethod
    def fromIsoWeekday(weekday):
        return Weekday(weekday - 1)

    @staticmethod
    def fromPythonWeekday(weekday):
        return Weekday(weekday)

    @staticmethod
    def fromGooglePlacesWeekday(weekday):
        result = weekday - 1
        return Weekday(result) if result >= 0 else Weekday(6)


class Hours:
    def __init__(self, open, close):
        self.open = open
        self.close = close


class Places:
    def __init__(self) -> None:
        self.gmaps = googlemaps.Client(
            key=lc.config["google"],
        )
        self.lat = 48.306284
        self.lng = 14.286215
        self.radius = 2000
        self.language = "de"

    def getPlaceInfo(self, place_id):
        return self.gmaps.place(
            place_id,
            fields=[
                "geometry/location",
                "formatted_address",
                "international_phone_number",
                "url",
                "name",
            ],
            language=self.language,
        ).get("result", {})

    def findPlace(self, query):
        res = self.gmaps.find_place(
            query,
            input_type="textquery",
            fields=["name", "place_id", "formatted_address"],
            location_bias=f"circle:{self.radius}@{self.lat},{self.lng}",
            language=self.language,
        )
        if res.get("status", None) == "OK":
            return res.get("candidates", [])
        return []

    def checkOpenHours(
        self,
        place_id,
        target_day: datetime.date,
        target_time=datetime.time(19, 0),
        duration=datetime.timedelta(hours=3),
    ):
        target_date = datetime.datetime.combine(target_day, target_time)

        place_detail = self.gmaps.place(place_id, fields=["opening_hours"])
        if place_detail["status"] != "OK" or len(place_detail["result"]) == 0:
            return (Openness.UNKNOWN, "")

        opening_hours = place_detail["result"]["opening_hours"]["periods"]
        # Returns something like this:
        # [
        #     {
        #         "close": {
        #             "day": 1,
        #             "time": "1500"
        #         },
        #         "open": {
        #             "day": 1,
        #             "time": "1200"
        #         }
        #     },
        #     {
        #         "close": {
        #             "day": 1,
        #             "time": "2230"
        #         },
        #         "open": {
        #             "day": 1,
        #             "time": "1700"
        #         }
        #     },
        #     {
        #         "close": {
        #             "day": 2,
        #             "time": "1500"
        #         },
        #         "open": {
        #             "day": 2,
        #             "time": "1200"
        #         }
        #     }
        # ...
        # ]

        wd = Weekday.fromPythonWeekday(
            target_day.weekday(),
        ).toGooglePlacesWeekday()

        # Get opening hour for tuesday == 2
        target_day_opening_hours = filter(
            lambda entry: entry["open"]["day"] == wd,
            opening_hours,
        )

        hours = []

        # TODO: Transform weekday and time to actual datetimes,
        #  for easier handling
        for oh in target_day_opening_hours:
            open = datetime.datetime.combine(
                target_day + datetime.timedelta(days=oh["open"]["day"] - wd),
                datetime.time(
                    int(oh["open"]["time"][0:2]),
                    int(oh["open"]["time"][2:4]),
                ),
            )
            close = datetime.datetime.combine(
                target_day + datetime.timedelta(days=oh["close"]["day"] - wd),
                datetime.time(
                    int(oh["close"]["time"][0:2]),
                    int(oh["close"]["time"][2:4]),
                ),
            )

            if open.date() < target_day:
                open += datetime.timedelta(7)

            if close.date() < target_day:
                close += datetime.timedelta(7)

            hours.append(Hours(open, close))

        weekday_text = "\n".join(
            place_detail["result"]["opening_hours"]["weekday_text"],
        )

        if len(hours) < 1:
            return (Openness.CLOSED, weekday_text)

        for h in hours:
            if h.open <= target_date and h.close > target_date:
                if h.close - target_date < duration:
                    return (Openness.SHORT, weekday_text)
                else:
                    return (Openness.OPEN, weekday_text)

        return (Openness.CLOSED, weekday_text)


if __name__ == "__main__":
    place = Places()
    while True:
        q = input("Search for location: ")
        r = place.findPlace(q)
        print(r)
        i = place.getPlaceInfo(r[0]["place_id"])
        print(i)
        print(
            place.checkOpenHours(
                r[0]["place_id"],
                datetime.date.fromisoformat("2023-05-16"),
            ),
        )



# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

import logging
import re
from datetime import date
from datetime import datetime
from enum import Enum

import mysql.connector

from leistungsbot import leistungs_config as lc
from leistungsbot.google_place import Places
from leistungsbot.leistungs_returns import LeistungsReturnCodes


class LeistungsTagState(Enum):
    NONE = 0
    OPEN = 1
    CLOSED = 2


class LeistungsDB:
    def __init__(self):
        self.google = Places()
        self.connect()

    def connect(self):
        try:
            self.mydb = mysql.connector.connect(
                host=lc.config["mysql"]["host"],
                database=lc.config["mysql"]["db"],
                user=lc.config["mysql"]["user"],
                password=lc.config["mysql"]["password"],
                ssl_disabled=True,
                collation="utf8mb4_general_ci",
            )
            return True
        except Exception as e:
            self.mydb = mysql.connector.MySQLConnection()
            logging.error("Error while connecting to MySQL", e)
            return False

    def convert(self, mysql_res, skinny_bitch=False):
        if not mysql_res:
            return None

        elif isinstance(mysql_res, dict):
            for k in mysql_res:
                mysql_res[k] = self.convert(mysql_res[k], skinny_bitch)

        elif type(mysql_res) in [tuple, list]:
            mysql_res = [self.convert(i, skinny_bitch) for i in mysql_res]

            if skinny_bitch and len(mysql_res) == 1:
                return mysql_res[0]

        elif isinstance(mysql_res, bytearray):
            pass

        elif isinstance(mysql_res, str):
            if "POINT" in mysql_res:
                return re.findall(r"[\d\.]+", mysql_res)

        return mysql_res

    def checkConnection(self):
        if self.mydb.is_connected():
            db_Info = self.mydb.get_server_info()
            print("Connected to MySQL Server version ", db_Info)
            cursor = self.mydb.cursor()
            cursor.execute("select database();")
            record = cursor.fetchone()
            print("You're connected to database: ", record)
        return self.mydb.is_connected()

    def addUser(self, user_id: int, chat_id: int = None):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        try:
            sql = """INSERT INTO `members` (`user_id`, `chat_id`, `score`, `joined`)
                VALUES (%s, %s, %s, %s);"""
            values = (
                user_id,
                chat_id,
                0,
                datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            )
            logging.debug(sql % values)
            cursor.execute(sql, values)
        except mysql.connector.IntegrityError:
            cursor.execute(
                "UPDATE `members` SET `left` = %s, `chat_id` = %s WHERE `user_id` = %s;",
                (None, chat_id, user_id),
            )
        self.mydb.commit()

    def getUsers(self):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT * FROM `members`;"
        values = ()
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getUserKey(self, user_id):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `key` FROM `members` WHERE `user_id` = %s;"
        values = (user_id,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getUserScore(self, user_id: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `score` FROM `members` WHERE `user_id` = %s;"
        values = (user_id,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def setUserScore(self, user_id: int, score: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "UPDATE `members` SET `score` = 'NULL', chat_id = %s WHERE `user_id` = %s;"
        values = (score, user_id)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def increaseUserScore(self, user_id: int, value: int):
        self.setUserScore(user_id, self.getUserScore(user_id) + value)

    def removeUser(self, user_id: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "UPDATE `members` SET `left` = %s WHERE `user_id` = %s;"
        # .strftime('%Y-%m-%d %H:%M:%S')
        values = (datetime.now(), user_id)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def addLocation(self, place_id: str, name: str) -> LeistungsReturnCodes:
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        info = self.google.getPlaceInfo(place_id)
        cursor = self.mydb.cursor()
        retry = True
        orig_name = name
        cnt = 1
        res = LeistungsReturnCodes.OK
        while retry:
            try:
                sql = "INSERT INTO `locations` (`name`, `google-place-id`, `lat`, `lng`, `address`, `phone`, `url`) VALUES (%s, %s, %s, %s, %s, %s, %s);"
                values = (
                    name,
                    place_id,
                    info.get("geometry", {})
                    .get("location", {})
                    .get("lat", None),
                    info.get("geometry", {})
                    .get(
                        "location",
                        {},
                    )
                    .get("lng", None),
                    info.get("formatted_address", None),
                    info.get("international_phone_number", None),
                    info.get("url", None),
                )
                logging.debug(sql % values)
                cursor.execute(sql, values)
                retry = False
            except mysql.connector.IntegrityError:
                location = self.getLocationInfo(name)
                if location.get("google-place-id") == place_id:
                    retry = False
                    res = LeistungsReturnCodes.DB_DUPLICATE
                else:
                    cnt += 1
                    name = orig_name + str(cnt)
        self.mydb.commit()
        return res

    def removeLocation(self, key):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "DELETE FROM `locations` WHERE `key` = %s;"
        values = (key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def getAllLocations(self):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `name` FROM `locations`;"
        values = ()
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getLocationKey(self, location_name):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `key` FROM `locations` WHERE `name` = %s;"
        values = (location_name,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getLocationName(self, location_key):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `name` FROM `locations` WHERE `key` = %s;"
        values = (location_key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getVisitedLocations(self):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `name`, `key` FROM `locations` WHERE `visited` = TRUE;"
        values = ()
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getVirgineLocations(self):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `name`, `key` FROM `locations` WHERE `visited` = FALSE ORDER BY `name` DESC;"
        values = ()
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getLocationInfo(self, location_name):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `locations` WHERE `name` = %s;"
        values = (location_name,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getLocationInfoByKey(self, key: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `locations` WHERE `key` = %s;"
        values = (key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def setLocationVisitedState(self, location_name, visited=True):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "UPDATE `locations` SET `visited` = %s WHERE `name` = %s;"
        values = (visited, location_name)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def setLocationVisitedStateKey(self, location_key, visited=True):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "UPDATE `locations` SET `visited` = %s WHERE `key` = %s;"
        values = (visited, location_key)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def rateLocation(self, location_name, user_id, rating):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        uKey = self.getUserKey(user_id)
        lKey = self.getLocationKey(location_name)
        cursor = self.mydb.cursor()
        sql = "INSERT INTO `location_rating` (`location`, `member`, `rating`) VALUES (%s, %s, %s);"
        values = (lKey, uKey, rating)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def rateLocationKey(self, location_key, user_id, rating):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        uKey = self.getUserKey(user_id)
        lKey = location_key
        cursor = self.mydb.cursor()
        sql = "INSERT INTO `location_rating` (`location`, `member`, `rating`) VALUES (%s, %s, %s);"
        values = (lKey, uKey, rating)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def getAvgLocationRating(self, location_name):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        lKey = self.getLocationKey(location_name)
        cursor = self.mydb.cursor()
        sql = "SELECT AVG(`rating`) FROM `location_rating` WHERE `location` = %s;"
        values = (lKey,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        res = self.convert(cursor.fetchall(), skinny_bitch=True)
        return res if res else 0

    def getUserLocationRating(self, location_name, user_id):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        uKey = self.getUserKey(user_id)
        lKey = self.getLocationKey(location_name)
        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT `rating` FROM `location_rating` WHERE `location` = %s AND `member` = %s;"
        values = (lKey, uKey)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getAvgUserLocationRating(self, user_id):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        uKey = self.getUserKey(user_id)
        cursor = self.mydb.cursor()
        sql = (
            "SELECT AVG(`rating`) FROM `location_rating` WHERE `member` = %s;"
        )
        values = (uKey,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def addLeistungsTag(
        self,
        date: datetime,
        location_name: str,
        poll_id: int,
        venue_id: int,
        type: int,
    ):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        lKey = self.getLocationKey(location_name)
        cursor = self.mydb.cursor()
        sql = "INSERT INTO `leistungstag` (`location`, `date`, `poll_id`, `venue_id`, `type`) VALUES (%s, %s, %s, %s, %s);"
        # .strftime('%Y-%m-%d')
        values = (lKey, date, poll_id, venue_id, type)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def getLeistungstag(self, key: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `leistungstag` WHERE `key` = %s;"
        values = (key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getLeistungstageByDate(self, date: date):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `leistungstag` WHERE DATE(`date`) = %s;"
        values = (date,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getLeistungsTagKeyPollId(self, poll_id: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `key` FROM `leistungstag` WHERE `poll_id` = %s;"
        values = (poll_id,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getOpenLeistungsTag(self, type: int = None):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        if type:
            sql = "SELECT * FROM `leistungstag` WHERE `type` = %s AND `closed` = %s ORDER BY `date`;"
            values = (int(type), False)
        else:
            sql = "SELECT * FROM `leistungstag` WHERE `closed` = %s ORDER BY `date`;"
            values = (False,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getClosedLeistungsTag(self, type: int = None):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        if type:
            sql = "SELECT * FROM `leistungstag` WHERE `type` = %s AND `closed` = %s ORDER BY `date`;"
            values = (int(type), True)
        else:
            sql = "SELECT * FROM `leistungstag` WHERE `closed` = %s ORDER BY `date`;"
            values = (True,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def closeLeistungstag(self, leistungstag_key: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "UPDATE `leistungstag` SET `closed` = %s WHERE `key` = %s;"
        values = (True, leistungstag_key)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def removeLeistungstag(self, key: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "DELETE FROM `leistungstag` WHERE `key` = %s;"
        values = (key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def getHistory(self, type: int = None, limit: int = 100):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = f"SELECT * FROM (SELECT * FROM `leistungstag` {'WHERE `type` = %s' if type else ''} ORDER BY `date` DESC LIMIT %s) SQ ORDER BY `date`;"
        if type:
            values = (int(type), int(limit))
        else:
            values = (int(limit),)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getHistoryCount(self, type: int = None):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        if type:
            sql = "SELECT COUNT(*) FROM `leistungstag` WHERE `type` = %s;"
            values = (int(type),)
        else:
            sql = "SELECT COUNT(*) FROM `leistungstag`;"
            values = ()
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchone(), True)

    def getLatest(self, type: int = None, state=LeistungsTagState.NONE):
        return self.getPrevious(type, state, 1)

    def getLeistungsTags(
        self,
        type: int = None,
        state=LeistungsTagState.NONE,
        max_results: int = 0,
        before: datetime = None,
    ):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        select = "SELECT * FROM `leistungstag`"
        where = ""
        values = ()
        order = "ORDER BY `date` DESC"

        if before:
            if len(where) > 0:
                where += " AND "
            else:
                where = "WHERE "
            where += "`date` <= %s"
            values += (before,)

        if type:
            if len(where) > 0:
                where += " AND "
            else:
                where = "WHERE "
            where += "`type` = %s"
            values += (int(type),)

        if state != LeistungsTagState.NONE:
            if len(where) > 0:
                where += " AND "
            else:
                where = "WHERE "
            where += "`closed` = %s"
            values += (state == LeistungsTagState.CLOSED,)

        limit = ""
        if max_results > 0:
            limit = "LIMIT %s"
            values += (max_results,)

        sql = f"{select} {where} {order} {limit};"
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getParticipants(self, leistungstag_key: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        sql = "SELECT `member` FROM `participants` WHERE `event` = %s;"
        values = (leistungstag_key,)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        return self.convert(cursor.fetchall())

    def getLatestParticipants(self, type=None):
        key = self.getLatest(type).get("key")
        return self.getParticipants(key)

    def addParticipant(self, user_id: int, poll_id: int):
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        user_key = self.getUserKey(user_id)
        leistungstag_key = self.getLeistungsTagKeyPollId(poll_id)
        cursor = self.mydb.cursor()
        sql = "INSERT INTO `participants` (`member`, `event`) VALUES (%s, %s);"
        values = (user_key, leistungstag_key)
        logging.debug(sql % values)
        cursor.execute(sql, values)
        self.mydb.commit()

    def getMostRecentLeistungstag(self) -> dict:
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `leistungstag` ORDER BY `date` DESC LIMIT 1"
        cursor.execute(sql)
        return self.convert(cursor.fetchone(), True)

    def getLeistungstagByNumber(self, number: int) -> dict | None:
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor(dictionary=True)
        sql = "SELECT * FROM `leistungs_view` WHERE number = %s"
        cursor.execute(sql, (number,))
        return self.convert(cursor.fetchone(), True)

    def switchLeistungstagLocation(
        self,
        lt_id: int,
        old_location_id: int,
        new_location_id: int,
    ) -> None:
        if not self.mydb.is_connected():
            if not self.connect():
                logging.error("No connection to DataBase possible")
                raise Exception("No connection to DataBase available")

        cursor = self.mydb.cursor()
        update_lt_sql = (
            "UPDATE `leistungstag` SET `location` = '%s' WHERE `key` = '%s'"
        )
        update_old_location = "UPDATE `locations` as l SET `visited` = (SELECT count(*) FROM leistungstag WHERE location = l.`key` LIMIT 1) WHERE `key` = '%s'"
        update_new_location = (
            "UPDATE `locations` as l SET `visited` = 1 WHERE `key` = '%s'"
        )

        cursor.execute(update_lt_sql, (new_location_id, lt_id))
        cursor.execute(update_old_location, (old_location_id,))
        cursor.execute(update_new_location, (new_location_id,))
        self.mydb.commit()


if __name__ == "__main__":
    logging.basicConfig(filename="myapp.log", level=logging.DEBUG)
    db = LeistungsDB()
    db.checkConnection()
    # db.addUser(4711)
    # db.addLocation('ChIJ5UvV55IHbUcRMq6el31MzZI', 'Cafe Phönixhof')
    # db.addLeistungsTag(datetime.now(), 'Cafe Phönixhof', 42069)
    # db.addParticipant(4711, 42069)
    a = db.getOpenLeistungsTag()
    print(a)



from __future__ import annotations

import argparse
import os

import confuse

template = {
    "mysql": {
        "host": confuse.Optional(str),
        "db": confuse.Optional(str),
        "user": confuse.Optional(str),
        "password": confuse.Optional(str),
    },
    "google": str,
    "bot_token": str,
    "api_hash": str,
    "api_id": [int, str],
    "chat_id": [int, str],
    "leistungschat_id": [int, str],
    "leistungsadmin_id": [int, str],
    "usernames": confuse.StrSeq(),
}

__config = confuse.Configuration("LeistungsBot", __name__)
if os.environ.get("LEISTUNGSBOT_CONFIG_FILE"):
    __config.set_file(
        os.environ.get(
            "LEISTUNGSBOT_CONFIG_FILE",
        ),
        base_for_paths=True,
    )
__config.set_env()
try:
    config = __config.get(template)  # NOQA
except Exception:
    print(
        "Configuration error: "
        "Please make sure to set the environment variables or provide a configuration file.",
    )
    config = None


def set_args(namespace: argparse.Namespace, dots: bool = False):
    if namespace.config:
        __config.set_file(
            namespace.config,
            base_for_paths=True,
        )
    __config.set_args(namespace, dots)
    global config
    config = __config.get(template)  # NOQA


def __getitem__(self, key):
    return config.get(key)


if __name__ == "__main__":
    assert config["mysql"]["user"] == "leistungs_user"



# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

import threading
import time
from datetime import datetime
from datetime import timedelta

import schedule
import telebot

from leistungsbot import leistungs_config as lc
from leistungsbot.BotHelper import LeistungsTyp
from leistungsbot.leistungs_db import LeistungsDB
from leistungsbot.leistungs_db import LeistungsTagState


class Scheduler:
    def __init__(self, bot: telebot.TeleBot) -> None:
        self.bot = bot
        self.db = LeistungsDB()
        self.schedule = schedule
        self.schedule.every().day.at("12:00").do(self.send_reminder)
        self.schedule.every().monday.at("12:00").do(self.send_reservation)
        self.schedule.every().day.at("19:00").do(self.close_previous)
        self.start()

    def run_continuously(self, interval=1):
        """Continuously run, while executing pending jobs at each
        elapsed time interval.
        @return cease_continuous_run: threading. Event which can
        be set to cease continuous run. Please note that it is
        *intended behavior that run_continuously() does not run
        missed jobs*. For example, if you've registered a job that
        should run every minute and you set a continuous run
        interval of one hour then your job won't be run 60 times
        at each interval but only once.
        """
        cease_continuous_run = threading.Event()

        class ScheduleThread(threading.Thread):
            @classmethod
            def run(cls):
                while not cease_continuous_run.is_set():
                    self.schedule.run_pending()
                    time.sleep(interval)

        continuous_thread = ScheduleThread()
        continuous_thread.start()
        return cease_continuous_run

    def start(self):
        # Start the background thread
        self.stop_run_continuously = self.run_continuously()

    def stop(self):
        # Stop the background thread
        self.stop_run_continuously.set()

    def send_reminder(self, type: LeistungsTyp = None):
        polls = self.db.getOpenLeistungsTag(type)
        for poll in polls:
            if (datetime.now() + timedelta(days=2)).date() == poll["date"]:
                self.bot.send_message(
                    lc.config["leistungschat_id"],
                    "Reminder. Übermorgen ises so weit, daun is endlich wieder Leistungstag.",
                    reply_to_message_id=poll["poll_id"],
                )

    def send_reservation(self, type: LeistungsTyp = None):
        polls = self.db.getClosedLeistungsTag(type)
        for poll in polls:
            if (datetime.now() + timedelta(days=1)).date() == poll["date"]:
                info = self.db.getLocationInfo(
                    self.db.getLocationName(poll["location"]),
                )
                self.bot.forward_message(
                    lc.config["leistungsadmin_id"],
                    lc.config["leistungschat_id"],
                    poll["poll_id"],
                )
                self.bot.send_message(
                    lc.config["leistungsadmin_id"],
                    f'Host eam e scho für moagn schiach reserviert ({info["phone"]})? {info["url"]}',
                )

    def close_previous(self, type: LeistungsTyp = None):
        previous = self.db.getLeistungsTags(
            type,
            LeistungsTagState.OPEN,
            before=datetime().now(),
        )
        for prev in previous:
            location = self.db.getLocationName(prev["location"])
            self.bot.send_message(
                lc.config["leistungsadmin_id"],
                f"Schaut so aus ois ob do a alter Leistungstag nuned geclosed worden is...I schlias {location} fia eich.",
            )
            self.helper.db.closeLeistungstag(prev["key"])
            self.bot.stop_poll(
                lc.config["leistungschat_id"],
                prev["poll_id"],
            )


if __name__ == "__main__":
    s = Scheduler(telebot.TeleBot(lc.config["bot_token"]))
    s.start()



# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

from leistungsbot.Bot import main as Bot

Bot()



# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

import json
import os
import pickle
import random
import tempfile
from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from enum import IntEnum

import telebot
from telebot.types import InlineKeyboardButton
from telebot.types import InlineKeyboardMarkup
from telebot.types import KeyboardButton
from telebot.types import ReplyKeyboardMarkup
from telegram_bot_calendar import DetailedTelegramCalendar
from telegram_bot_calendar import LSTEP

from leistungsbot import leistungs_config as lc
from leistungsbot.google_place import Places
from leistungsbot.leistungs_db import LeistungsDB
from leistungsbot.leistungs_returns import LeistungsReturnCodes


class LeistungsTyp(IntEnum):
    NORMAL = 1
    KONKURENZ = 2
    ZUSATZ = 3


class Helper:
    def __init__(self, bot: telebot.TeleBot) -> None:
        self.bot = bot
        self.db = LeistungsDB()
        self.temp_dir = tempfile.gettempdir()
        self.google = Places()
        self.dateformat = "%d.%m.%Y"

    def filter(self):
        def inn(callback):
            try:
                data = json.loads(callback.data)
                cmd = [*data][0]
                return cmd.startswith("🍻")
            except BaseException:
                return False

    def escape_markdown(self, text: str, markdown_version: int = 2):
        return telebot.formatting.escape_markdown(text)

    def get_full_temp_file(self, file_id: str):
        return os.path.join(self.temp_dir, str(file_id) + "_leistung")

    def get_full_temp_file_handle(self, file_id: str, rw=False):
        path = self.get_full_temp_file(file_id)
        if rw:
            return open(path, "wb")
        else:
            return open(path, "rb")

    def store_to_rand_file(self, data):
        rand_id = random.randint(10000, 100000)
        with self.get_full_temp_file_handle(rand_id, True) as handle:
            pickle.dump(data, handle)
        return rand_id

    def peak_from_rand_file(self, rand_id):
        with self.get_full_temp_file_handle(rand_id) as handle:
            return pickle.load(handle)

    def load_from_rand_file(self, rand_id):
        with self.get_full_temp_file_handle(rand_id) as handle:
            data = pickle.load(handle)
        os.remove(self.get_full_temp_file(rand_id))
        return data

    def location_keyboard(self):
        markup = ReplyKeyboardMarkup(row_width=1, one_time_keyboard=True)
        for location in self.db.getVirgineLocations():
            markup.add(KeyboardButton(location[0]))
        return markup

    def rating_keyboard(self):
        markup = ReplyKeyboardMarkup(row_width=1, one_time_keyboard=True)
        # hack to allow for .25 rating
        for i in range(0, 600, 25):
            markup.add(KeyboardButton(self.get_stars(i / 100)))
            if i == 500:
                return markup
        return markup

    def virgine_location_button(self):
        markup = InlineKeyboardMarkup(row_width=1)
        for location in self.db.getVirgineLocations():
            markup.add(
                InlineKeyboardButton(
                    location[0],
                    callback_data=json.dumps({"🍻location": location[1]}),
                ),
            )
        return markup

    def polls_button(self, leistungstage, additional_button=False):
        markup = InlineKeyboardMarkup(row_width=1)
        if leistungstage:
            for leistungstag in leistungstage:
                callback_key = (
                    "🍻closed" if leistungstag["closed"] == 1 else "🍻open"
                )
                markup.add(
                    InlineKeyboardButton(
                        self.db.getLocationName(leistungstag["location"]),
                        callback_data=json.dumps(
                            {callback_key: leistungstag["key"]},
                        ),
                    ),
                )
        if additional_button:
            markup.add(
                InlineKeyboardButton(
                    "Nö",
                    callback_data=json.dumps({"🍻no": None}),
                ),
            )
        return markup

    def open_polls_button(self, additional_button=False):
        leistungstage = self.db.getOpenLeistungsTag()
        return self.polls_button(leistungstage, additional_button)

    def confirm_leistungstag_button(self, leistungstag_key):
        """Use for sendreminder or closepoll"""
        markup = InlineKeyboardMarkup(row_width=2)
        markup.add(
            InlineKeyboardButton(
                "Na",
                callback_data=json.dumps({"🍻cancle": None}),
            ),
            InlineKeyboardButton(
                "Jo",
                callback_data=json.dumps(
                    {"🍻open": leistungstag_key},
                ),
            ),
        )
        return markup

    def check_open_hours_keyboard(
        self,
        agree_message,
        abort_message="Hoitaus. Abort!",
    ):
        markup = InlineKeyboardMarkup(row_width=1)
        markup.add(
            InlineKeyboardButton(
                agree_message,
                callback_data=json.dumps({"🍻open_hours_checked": True}),
            ),
        )
        markup.add(
            InlineKeyboardButton(
                abort_message,
                callback_data=json.dumps({"🍻open_hours_checked": False}),
            ),
        )
        return markup

    def search_location(self, query):
        g_places = self.google.findPlace(query)
        return len(g_places), self.store_to_rand_file(g_places)

    def search_location_button(self, rand_id):
        g_places = self.peak_from_rand_file(rand_id)
        markup = InlineKeyboardMarkup(row_width=1)
        for i in range(len(g_places)):
            markup.add(
                InlineKeyboardButton(
                    f"""{g_places[i]['name']} - {
                        g_places[i]
                        ['formatted_address']
                    }""",
                    callback_data=json.dumps({"🍻search": (rand_id, i)}),
                ),
            )
        return markup

    def restore_search_location_button(self, rand_id):
        markup = InlineKeyboardMarkup(row_width=1)
        g = self.peak_from_rand_file(rand_id)
        for i in range(len(g)):
            markup.add(
                InlineKeyboardButton(
                    g[i]["name"],
                    callback_data=json.dumps({"🍻search": (rand_id, i)}),
                ),
            )
        return markup

    def approve_location_button(self, rand_id, index):
        markup = InlineKeyboardMarkup(row_width=1)
        markup.add(
            InlineKeyboardButton(
                "Ned mei location",
                callback_data=json.dumps({"🍻select": (rand_id, -1)}),
            ),
            InlineKeyboardButton(
                "Des is mei location",
                callback_data=json.dumps(
                    {"🍻select": (rand_id, index)},
                ),
            ),
        )
        return markup

    def leistungstag_type_button(self, key: str):
        markup = InlineKeyboardMarkup(row_width=1)
        for i in LeistungsTyp:
            markup.add(
                InlineKeyboardButton(
                    i.name,
                    callback_data=json.dumps({key: i}),
                ),
            )
        return markup

    def leistungstag_poll_type_button(self):
        return self.leistungstag_type_button("🍻poll_type")

    def leistungstag_history_type_button(self):
        return self.leistungstag_type_button("🍻history_type")

    def leistungstag_purge_type_button(self):
        return self.leistungstag_type_button("🍻purge_type")

    def leistungstag_button(self, key: str, type: int, limit: int = 100):
        history = self.db.getHistory(type, limit)
        markup = InlineKeyboardMarkup(row_width=1)
        for i in range(len(history)):
            name = (
                str(i + 1)
                + "."
                + str(self.db.getLocationName(history[i]["location"]))
            )
            markup.add(
                InlineKeyboardButton(
                    name,
                    callback_data=json.dumps({key: history[i]["key"]}),
                ),
            )
        return markup

    def leistungstag_history_button(self, type):
        return self.leistungstag_button("🍻history", type)

    def leistungstag_dry_purge_button(self, type):
        return self.leistungstag_button("🍻dry_purge", type, 5)

    def leistungstag_purge_button(self, key):
        markup = InlineKeyboardMarkup(row_width=2)
        markup.add(
            InlineKeyboardButton(
                "Des mochn ma ned!",
                callback_data=json.dumps({"🍻cancle": None}),
            ),
            InlineKeyboardButton(
                "Weg damit",
                callback_data=json.dumps(
                    {"🍻purge": key},
                ),
            ),
        )
        return markup

    def unkown_location_button(self, location):
        markup = InlineKeyboardMarkup(row_width=2)
        markup.add(
            InlineKeyboardButton(
                "Na",
                callback_data=json.dumps({"🍻cancle": None}),
            ),
            InlineKeyboardButton(
                "Jo",
                callback_data=json.dumps(
                    {"🍻q": location},
                ),
            ),
        )
        return markup

    def dry_run_button(self, rand_id):
        markup = InlineKeyboardMarkup(row_width=2)
        markup.add(
            InlineKeyboardButton(
                "Na",
                callback_data=json.dumps({"🍻cancle": None}),
            ),
            InlineKeyboardButton(
                "Jo",
                callback_data=json.dumps(
                    {"🍻publish": rand_id},
                ),
            ),
        )
        return markup

    def sender_has_permission(self, msg):
        sender = self.bot.get_chat_member(
            lc.config["leistungschat_id"],
            msg.from_user.id,
        )
        return sender.status == "administrator" or sender.status == "creator"

    def send_nude(self, chat_id):
        gif = "https://cdn.porngifs.com/img/%s" % (random.randint(1, 39239))
        self.bot.send_animation(
            chat_id,
            gif,
            caption="brought to you by Maxmaier",
            has_spoiler=True,
        )

    def next_leistungstag(self):
        return datetime.now() + timedelta(
            days=(8 - datetime.now().weekday()),
        )

    def send_leistungstag(
        self,
        chat_id,
        location: str,
        type: LeistungsTyp = LeistungsTyp.NORMAL,
        date: datetime = None,
        dry_run: bool = True,
    ):
        if not date:
            date = self.next_leistungstag()
        date_str = date.strftime(self.dateformat)
        # close_date = date - timedelta(hours=12)
        info = self.db.getLocationInfo(location)
        venue_id = self.bot.send_venue(
            chat_id,
            latitude=info["lat"],
            longitude=info["lng"],
            title=info["name"],
            address=info["address"],
        )
        count = self.db.getHistoryCount(type)
        count = count + 1 if count else 1
        if type == LeistungsTyp.NORMAL:
            question = f'Leistungstag {count}: am {date_str} in "{location}"'
        elif type == LeistungsTyp.KONKURENZ:
            question = f'Konkurrenz Leistungstag {count}: am {date_str} in "{location}"'
        elif type == LeistungsTyp.ZUSATZ:
            question = f'Leistungstag Zusatztermin {count}: am {date_str} in "{location}"'
        else:
            question = "Keine Ahnung wos wia grad polln..."
        poll_message = self.bot.send_poll(
            chat_id,
            question,
            ["Bin dabei", "Keine Zeit"],
            allows_multiple_answers=False,
            explanation="Soi i da jez a nu erklährn wie ma obstimmt?",
            is_anonymous=False,
        )
        if dry_run:
            rand_id = self.store_to_rand_file((location, type, date))
            self.bot.send_message(
                chat_id,
                "Woin ma des so veröffentlichen?",
                reply_markup=self.dry_run_button(rand_id),
            )
        else:
            self.db.addLeistungsTag(
                date,
                location,
                poll_message.message_id,
                venue_id.message_id,
                int(type),
            )
            self.db.setLocationVisitedState(location, True)
            self.bot.pin_chat_message(chat_id, poll_message.message_id)

    def send_location_info(self, chat_id, place_id, reply_markup=None):
        info = self.google.getPlaceInfo(place_id)
        self.bot.send_venue(
            chat_id,
            info["geometry"]["location"]["lat"],
            info["geometry"]["location"]["lng"],
            info["name"],
            info["formatted_address"],
            google_place_id=place_id,
            reply_markup=reply_markup,
        )

    def approve_location(self, chat_id, rand_id, index):
        data = self.peak_from_rand_file(rand_id)
        self.send_location_info(
            chat_id,
            data[index]["place_id"],
            self.approve_location_button(rand_id, index),
        )

    def add_location(self, rand_id, index) -> LeistungsReturnCodes:
        data = self.load_from_rand_file(rand_id)
        return self.db.addLocation(
            data[index]["place_id"],
            data[index]["name"],
        )

    def remove_location(self, locationname):
        key = self.db.getLocationKey(locationname)
        self.db.removeLocation(key)

    def publish_leistungstag(self, rand_id):
        vals = self.load_from_rand_file(rand_id)
        self.send_leistungstag(
            lc.config["leistungschat_id"],
            vals[0],
            vals[1],
            vals[2],
            False,
        )

    def get_rand_len(self, rand_id):
        return len(self.peak_from_rand_file(rand_id))

    def get_stars(self, rating: float):
        if rating > 5:
            rating = 5
        res = "🌕" * int(rating)
        rating -= int(rating)
        if rating >= 0.625:
            res += "🌖"
        elif rating >= 0.325:
            res += "🌗"
        elif rating > 0:
            res += "🌘"
        return res.ljust(5, "🌑")

    def get_rating(self, rating: str):
        return (
            rating.count("🌕")
            + rating.count("🌖") * 0.75
            + rating.count("🌗") * 0.5
            + rating.count("🌘") * 0.25
        )

    def send_history_info(self, chat_id, leistungstag_key: int):
        ld = self.db.getLeistungstag(leistungstag_key)
        info = self.db.getLocationInfoByKey(ld["location"])
        rating = self.db.getAvgLocationRating(info["name"])

        message = (
            f"""*{self.escape_markdown(info.get('name'))}*"""
            + self.escape_markdown(
                f"""
{ld['date'].strftime(self.dateformat)}
{self.get_stars(rating)}
{info.get('address')}
{info.get('phone')}
{info.get('url')}""",
            )
        )

        self.bot.send_message(
            chat_id,
            message,
            parse_mode="MarkdownV2",
        )

    def send_location_info2(self, chat_id, location_key: int):
        info = self.db.getLocationInfoByKey(location_key)
        self.send_location_info(chat_id, info["google-place-id"])

        message = (
            f"""*{self.escape_markdown(info.get('name'))}*"""
            + self.escape_markdown(
                f"""
{info.get('address')}
{info.get('phone')}
{info.get('url')}""",
            )
        )
        self.bot.send_message(
            chat_id,
            message,
            parse_mode="MarkdownV2",
        )

    def send_purge_info(self, chat_id, leistungstag_key: int):
        ld = self.db.getLeistungstag(leistungstag_key)
        info = self.db.getLocationInfoByKey(ld["location"])
        rating = self.db.getAvgLocationRating(info["name"])

        message = (
            f"""*{self.escape_markdown(info.get('name'))}*"""
            + self.escape_markdown(
                f"""
{ld['date'].strftime(self.dateformat)}
{self.get_stars(rating)}
{info.get('address')}
{info.get('phone')}
{info.get('url')}""",
            )
        )

        self.bot.send_message(
            chat_id,
            message,
            parse_mode="MarkdownV2",
            reply_markup=self.leistungstag_purge_button(leistungstag_key),
        )

    def purge_leistungstag(self, leistungstag_key: int):
        ld = self.db.getLeistungstag(leistungstag_key)
        res = False
        try:
            res = self.bot.delete_message(
                lc.config["leistungschat_id"],
                ld["poll_id"],
            )
        except BaseException:
            pass
        try:
            res &= self.bot.delete_message(
                lc.config["leistungschat_id"],
                ld["venue_id"],
            )
        except BaseException:
            pass
        self.db.setLocationVisitedStateKey(ld["location"], False)
        self.db.removeLeistungstag(leistungstag_key)
        return res

    def excep(self, msg, error):
        self.bot.send_message(
            lc.config["chat_id"],
            f"""Error From Poll Bot!

Error  :: {error}

--------------------------------

Command:: {msg.text}

--------------------------------

UserDetails: {msg.from_user}

--------------------------------

Date   :: {msg.date}

--------------------------------

The Complete Detail:
{msg}


""",
        )

        return self.bot.reply_to(
            self.message,
            f"""An Unexpected Error Occured!
Error::  {error}
The error was informed to @eckphi""",
        )

    def pick_date(self, chat_id):
        calendar, step = DetailedTelegramCalendar(
            min_date=date.today(),
        ).build()
        self.bot.send_message(
            chat_id,
            f"Select {LSTEP[step]}",
            reply_markup=calendar,
        )

    def date_suggester(self):
        markup = InlineKeyboardMarkup(row_width=1)
        next_tuseday = self.next_leistungstag()
        markup.add(
            InlineKeyboardButton(
                next_tuseday.strftime(self.dateformat),
                callback_data=json.dumps(
                    {"🍻poll_date": next_tuseday.strftime(self.dateformat)},
                ),
            ),
        )
        markup.add(
            InlineKeyboardButton(
                "Anderes Datum",
                callback_data=json.dumps({"🍻poll_date": None}),
            ),
        )
        return markup

    def check_open_hours(
        self,
        location: str,
        date: date,
        time: time = time(19, 00),
    ):
        place_info = self.db.getLocationInfo(location)
        return self.google.checkOpenHours(
            place_info["google-place-id"],
            date,
            time,
        )


class PersistantLeistungsTagPoller:
    def __init__(
        self,
        helper: Helper,
        chat_id,
        location: str,
        type: LeistungsTyp = LeistungsTyp.NORMAL,
        date: datetime = None,
    ):
        self.chat_id = chat_id
        self.location = location
        self.type = type
        self.date = date if date else helper.next_leistungstag()
        self.helper = helper

    def dry_send_with_date(self, date: datetime):
        self.helper.send_leistungstag(
            self.chat_id,
            self.location,
            self.type,
            date,
            True,
        )

    def dry_send(self):
        self.helper.send_leistungstag(
            self.chat_id,
            self.location,
            self.type,
            self.date,
            True,
        )



# #############################################################################
#  "THE BEER-WARE LICENSE" (Revision 42):                                     #
#  @eckphi wrote this file. As long as you retain this notice you             #
#  can do whatever you want with this stuff. If we meet some day, and you think
#  this stuff is worth it, you can buy me a beer in return Poul-Henning Kamp  #
# #############################################################################
from __future__ import annotations

import argparse
import importlib.resources
import json
import logging
import time
from datetime import date
from datetime import datetime
from pathlib import Path
from typing import TypedDict

import telebot
from telebot import custom_filters
from telebot.handler_backends import State
from telebot.handler_backends import StatesGroup
from telebot.storage import StateMemoryStorage
from telegram_bot_calendar import DetailedTelegramCalendar
from telegram_bot_calendar import LSTEP

from leistungsbot import leistungs_config as lc
from leistungsbot.BotHelper import Helper
from leistungsbot.BotHelper import LeistungsTyp
from leistungsbot.BotHelper import PersistantLeistungsTagPoller
from leistungsbot.BotScheduler import Scheduler
from leistungsbot.google_place import Openness
from leistungsbot.leistungs_returns import LeistungsReturnCodes
from leistungsbot.package import _version

# States storage
# Now, you can pass storage to bot.
state_storage = StateMemoryStorage()  # you can init here another storage

#      ┌──────────────────────────────────────────────────────────┐
#      │                        HELP TEXT                         │
#      └──────────────────────────────────────────────────────────┘
"""
User Available Commands:
    1.  /leistungspoll
    2.  /add_location
    3.  /help
    4.  /purge
    5.  /mario
    6.  /start
    7.  /sendnudes
    8.  /rate_location
    9.  /show_participants
    10. /reminde_me
    11. /show_locations
    12. /remove_location
    13. /location_info
    14. /zusatzpoll
    15. /konkurrenzpoll
    16. /version

Developer Commands: #NOTE: ONLY @eckphi is
 allowed for these comands:
    1. /showIds
    2. /botlogs
"""

logger = telebot.logger
telebot.logger.setLevel(logging.DEBUG)  # Outputs debug messages to console.

# ──────────────────────────────────── States Group ──────────────────────


class LeistungsState(StatesGroup):
    # Just name variables differently
    normalLocation = (
        State()
    )  # creating instances of State class is enough from now
    konkurrenzLocation = State()
    zusatzLocation = State()
    searchLocation = State()
    purgeLeistungstag = State()
    historyLeistungstag = State()
    remindePoll = State()
    closePoll = State()
    sneakyClosePoll = State()
    removeLocation = State()
    rateLocation = State()
    genericLeistungsmessage = State()
    switcherooLeistungstagNumber = State()
    switcherooAlternateLocation = State()


class UserContext(TypedDict):
    leistungstag: dict | None


class LeistungsBot:
    def __init__(self) -> None:
        self.bot = bot = telebot.TeleBot(lc.config["bot_token"])
        self.bot.add_custom_filter(custom_filters.StateFilter(self.bot))
        self.helper = Helper(self.bot)
        self.scheduler = Scheduler(self.bot)
        self.poller = None
        self.last_text_nudes = datetime.min
        self.user_context: dict[int, UserContext] = {}

        @bot.callback_query_handler(func=DetailedTelegramCalendar.func())
        def cal(call):
            result, key, step = DetailedTelegramCalendar(
                min_date=date.today(),
            ).process(call.data)
            if not result and key:
                self.helper.bot.edit_message_text(
                    f"Select {LSTEP[step]}",
                    call.message.chat.id,
                    call.message.message_id,
                    reply_markup=key,
                )
            elif result:
                self.helper.bot.edit_message_text(
                    f"You selected {result}",
                    call.message.chat.id,
                    call.message.message_id,
                )
                if not self.poller:
                    self.helper.bot.send_message(
                        call.message.chat_id,
                        "Da is wohl was schiefglaufen, i kann ka poll findn...",
                    )
                    return
                if (
                    self.poller.type == LeistungsTyp.NORMAL
                    or self.poller.type == LeistungsTyp.KONKURENZ
                ) and result.weekday() != 1:
                    self.helper.bot.send_message(
                        call.message.chat.id,
                        "Blasphemie, des is ka Dienstag wast da du do ausgsuacht hast...alles auf eigene Gefahr!",
                    )
                    time.sleep(1)

                self.check_open_hours_before_sending(call, result)

        @bot.callback_query_handler(func=self.helper.filter())
        def callback_query(call):
            try:
                data = json.loads(call.data)
                if len(data) == 0:
                    bot.answer_callback_query(
                        call.id,
                        "SHHEEEEESH des hod ned funktioniert",
                    )
                    return
                cmd = [*data][0].replace("🍻", "")
                val = [*data.values()][0]
                bot.answer_callback_query(call.id, "Copy that")
                if cmd == "search":
                    self.helper.approve_location(
                        call.message.chat.id,
                        val[0],
                        val[1],
                    )
                elif cmd == "select":
                    if val[1] < 0:
                        if (self.helper.get_rand_len(val[0])) == 1:
                            self.bot.send_message(
                                call.message.chat.id,
                                "Daun füg a boa mehr infos zu deiner Suche dazua...",
                            )
                        else:
                            self.bot.send_message(
                                call.message.chat.id,
                                "Daun probiern mas numoi...",
                                reply_markup=self.helper.restore_search_location_button(
                                    val[0],
                                ),
                            )
                    else:
                        res = self.helper.add_location(val[0], val[1])
                        if res == LeistungsReturnCodes.DB_DUPLICATE:
                            self.bot.send_message(
                                call.message.chat.id,
                                "Des isch scho drin, du deppata!",
                            )
                elif cmd == "cancle":
                    self.process_cancle(call.message)
                elif cmd == "publish":
                    self.helper.publish_leistungstag(val)
                    bot.send_message(
                        call.message.chat.id,
                        "Hauma so veröffentlicht",
                    )
                elif cmd == "q":
                    self.process_search_location(call.message.chat.id, val)
                elif cmd == "history_type":
                    self.bot.send_message(
                        call.message.chat.id,
                        "Welchen Leistungstag willst da anschaun?",
                        reply_markup=self.helper.leistungstag_history_button(
                            LeistungsTyp(val),
                        ),
                    )
                elif cmd == "purge_type":
                    self.bot.send_message(
                        call.message.chat.id,
                        "Welchen Leistungstag willst löschen?",
                        reply_markup=self.helper.leistungstag_dry_purge_button(
                            LeistungsTyp(val),
                        ),
                    )
                elif cmd == "history":
                    self.helper.send_history_info(
                        call.message.chat.id,
                        val,
                    )
                elif cmd == "dry_purge":
                    self.helper.send_purge_info(
                        call.message.chat.id,
                        val,
                    )
                elif cmd == "purge":
                    if self.helper.purge_leistungstag(val):
                        bot.send_message(
                            call.message.chat.id,
                            "zack und weg ises",
                        )
                    else:
                        bot.send_message(
                            call.message.chat.id,
                            "De Nachrichtn muast leida manuell löschen",
                        )
                elif cmd == "location":
                    self.helper.send_location_info2(call.message.chat.id, val)
                elif cmd == "open":
                    if (
                        self.bot.get_state(
                            call.from_user.id,
                            call.message.chat.id,
                        )
                        == LeistungsState.remindePoll.name
                    ):
                        self.process_reminder(call.message, val)
                    elif (
                        self.bot.get_state(
                            call.from_user.id,
                            call.message.chat.id,
                        )
                        == LeistungsState.closePoll.name
                    ):
                        self.process_closepoll(call.message, val)
                    elif (
                        self.bot.get_state(
                            call.from_user.id,
                            call.message.chat.id,
                        )
                        == LeistungsState.sneakyClosePoll.name
                    ):
                        self.process_closepoll(call.message, val, True)
                    elif (
                        self.bot.get_state(
                            call.from_user.id,
                            call.message.chat.id,
                        )
                        == LeistungsState.genericLeistungsmessage.name
                    ):
                        self.process_generic_leistungsmessage(
                            call.message.reply_to_message,
                            val,
                        )
                elif cmd == "closed":
                    self.bot.reply_to(
                        call.message,
                        "Der Poll is scho closed. Willst wirklich on den reminden?",
                        reply_markup=self.helper.confirm_leistungstag_button(
                            val,
                        ),
                    )
                elif cmd == "no":
                    self.process_generic_leistungsmessage(
                        call.message.reply_to_message,
                        val,
                    )
                elif cmd == "poll_date":
                    if val:
                        if not self.poller:
                            self.helper.bot.send_message(
                                call.message.chat_id,
                                "Da is wohl was schiefglaufen, i kann ka poll findn...",
                            )
                        else:
                            self.check_open_hours_before_sending(
                                call,
                                datetime.strptime(
                                    val,
                                    self.helper.dateformat,
                                ).date(),
                            )
                    else:
                        self.helper.pick_date(call.message.chat.id)
                elif cmd == "open_hours_checked":
                    self.process_check_open_hours(call, val)
                else:
                    bot.send_message(
                        lc.config["chat_id"],
                        f"Hi Devs!!\nHandle this callback\n{cmd}",
                    )
                bot.edit_message_reply_markup(
                    call.message.chat.id,
                    call.message.message_id,
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(
                    call.message,
                    f"An error occurred!\nError: {error}",
                )
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["showIds"])
        def showIds(message):
            try:
                if message.from_user.username in lc.config["usernames"]:
                    file = open("joined_groups.txt", "r ")
                    bot.send_document(message.chat.id, file)
                    file.close()

            except Exception as error:
                bot.send_message(lc.config["chat_id"], str(error))

        @bot.message_handler(commands=["stats", "groups"])
        def stats(message):
            try:
                if message.from_user.username in lc.config["usernames"]:
                    print("Sending Stats To Owner")
                    with open("joined_groups.txt") as file:
                        group_ids = []
                        for line in file.readlines():
                            for group_id in line.split(" "):
                                group_ids.append(group_id)
                                no_of_polls = len(group_ids)
                                no_of_groups = len(list(set(group_ids)))
                        group_ids.clear()
                        bot.reply_to(
                            message,
                            f"Number of polls Made: {no_of_polls}\n#Nr of groups bot has been added to: {no_of_groups}",
                        )
                        file.close()
                else:
                    bot.reply_to(
                        message,
                        f"Sorry {message.from_user.username}! You Are Not Allowed To Use This Command,",
                    )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                try:
                    group_ids.clear()
                except BaseException:
                    pass
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["botlogs"])
        def ViewTheLogsFile(message):
            try:
                if message.from_user.username in lc.config["usernames"]:
                    print("Owner Asked For The Logs!")
                    file = open("POLL_LOGS.txt")
                    bot.send_document(
                        message.chat.id,
                        file,
                        timeout=60,
                        disable_notification=True,
                    )
                    file.close()
                    print("Logs Sent To Owner")
                else:
                    bot.reply_to(
                        message,
                        f"Sorry {message.from_user.username}! You Are Not Allowed For This Command.",
                    )
            except Exception as error:
                bot.reply_to(message, f"Error: {error}")

        @bot.message_handler(commands=["help"])
        def helper(message):
            return bot.reply_to(message, "Eiso i hüf da do ned...")

        @bot.message_handler(commands=["purge"])
        def purge(message):
            if not self.helper.sender_has_permission(message):
                bot.reply_to(
                    message,
                    "Diese Funktion ist nicht für den Pöbel gedacht.",
                )
                return
            try:
                self.process_purge(message)
            except IndexError:
                return bot.reply_to(
                    message,
                    f"""Lol!!! An error in the wild:
                    {message.text}

                    Which is invalid.
                    For more help use: /help
                    """,
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"""Error From Poll Bot!

                    Error  :: {error}

                    --------------------------------

                    Command:: {message.text}

                    --------------------------------

                    UserDetails: {message.from_user}

                    --------------------------------

                    Date   :: {message.date}

                    --------------------------------

                    The Complete Detail:
                    {message}


                    """,
                )

                return bot.reply_to(
                    message,
                    f"""An Unexpected Error Occured!
                    Error::  {error}
                    The error was informed to @eckphi""",
                )

        @bot.message_handler(commands=["alive"])
        def alive(message):
            bot.reply_to(
                message,
                f"Hey {message.from_user.username}, Ready To Serve You in version {_version.__version__}",
            )

        @bot.message_handler(commands=["start"])
        def start(message):
            bot.reply_to(
                message,
                f"Heya {message.from_user.username}, I am there to help you in polls. But this cmd is bit old try /help.",
            )

        @bot.message_handler(commands=["leistungspoll"])
        def poll_now(message):
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.normalLocation,
                    message.chat.id,
                )
                self.bot.send_message(
                    message.chat.id,
                    "Schick de nexte location muaz",
                    reply_markup=self.helper.location_keyboard(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["zusatzpoll"])
        def zusatz_poll(message):
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.zusatzLocation,
                    message.chat.id,
                )
                self.bot.send_message(
                    message.chat.id,
                    "Schick de nexte location muaz",
                    reply_markup=self.helper.location_keyboard(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["konkurrenzpoll"])
        def konkurrenz_poll(message):
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.konkurrenzLocation,
                    message.chat.id,
                )
                self.bot.send_message(
                    message.chat.id,
                    "Schick de nexte location muaz",
                    reply_markup=self.helper.location_keyboard(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["sendreminder"])
        def send_reminder(message: telebot.types.Message):
            try:
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.remindePoll,
                    message.chat.id,
                )

                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return

                # print(f'Message: {message.text}')
                command_parts = message.text.strip().split()

                if len(command_parts) == 1:
                    self.bot.reply_to(
                        message,
                        "An welchen Poll wüst reminden?",
                        reply_markup=self.helper.open_polls_button(),
                    )
                    return
                else:
                    target_date_str = command_parts[1]

                    try:
                        target_date = date.fromisoformat(target_date_str)
                    except BaseException:
                        target_date = None
                        self.bot.send_message(
                            message.chat.id,
                            f"Soi des a Datum sei? Schick ma wonn donn sowos wie {datetime.now().date().isoformat()}",
                        )
                        self.bot.reply_to(
                            message,
                            "An welchen Poll wüst reminden?",
                            reply_markup=self.helper.open_polls_button(),
                        )

                    # print(f'Date: {target_date}')

                    if target_date is not None:
                        successful = (
                            self.try_remind_to_leistungstag_on_a_specific_date(
                                message,
                                target_date,
                            )
                        )
                        if not successful:
                            self.bot.reply_to(
                                message,
                                "An welchen Poll wüst reminden?",
                                reply_markup=self.helper.open_polls_button(),
                            )

            except Exception as error:
                self.bot.delete_state(message.from_user.id, message.chat.id)

                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["closepoll"])
        def close_poll(message):
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return

                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.closePoll,
                    message.chat.id,
                )
                self.bot.reply_to(
                    message,
                    "Welchen Poll wüst closen?",
                    reply_markup=self.helper.open_polls_button(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["sneaky_closepoll"])
        def sneaky_close_poll(message: telebot.types.Message) -> None:
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return

                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.sneakyClosePoll,
                    message.chat.id,
                )
                self.bot.reply_to(
                    message,
                    "Welchen Poll wüst sneaky closen?",
                    reply_markup=self.helper.open_polls_button(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["sendnudes"])
        def send_nudes(message):
            try:
                if message.chat.type != "private":
                    bot.reply_to(
                        message,
                        "Bist deppad? Des is nix fürn Gruppen chat, du Drecksau.",
                    )
                else:
                    self.process_send_nudes(message.chat.id)
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["add_location"])
        def add_location(message):
            try:
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.searchLocation,
                    message.chat.id,
                )
                self.bot.send_message(
                    message.chat.id,
                    "Schick dei location idee muaz",
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["remove_location"])
        def remove_location_handler(message):
            try:
                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return

                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.removeLocation,
                    message.chat.id,
                )
                self.bot.reply_to(
                    message,
                    "Welche Location willst löschen?",
                    reply_markup=self.helper.location_keyboard(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["history"])
        def history(message):
            try:
                self.process_history(message)
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["rate_location"])
        def rate_location_handler(message):
            try:
                if message.chat.type != "private":
                    self.helper.bot.reply_to(
                        message,
                        "Und wenn ma des ned im Gruppenchat machen, du Bauernschädl?",
                    )
                else:
                    leistungstag = self.helper.db.getLeistungsTags(
                        LeistungsTyp.NORMAL,
                        max_results=1,
                        before=datetime.now(),
                    )[0]
                    self.helper.send_location_info2(
                        message.chat.id,
                        leistungstag["location"],
                    )
                    self.helper.bot.send_message(
                        message.chat.id,
                        "Wiafü Monde wüst erm geben?",
                        reply_markup=self.helper.rating_keyboard(),
                    )
                    self.bot.set_state(
                        message.from_user.id,
                        LeistungsState.rateLocation,
                        message.chat.id,
                    )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["show_locations"])
        def show_locations(message):
            try:
                bot.reply_to(
                    message,
                    "Des san de nächsten Locations",
                    reply_markup=self.helper.virgine_location_button(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(commands=["message"])
        def message_handler(message: telebot.types.Message):
            try:
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.remindePoll,
                    message.chat.id,
                )

                if not self.helper.sender_has_permission(message):
                    self.bot.reply_to(
                        message,
                        "Diese Funktion ist nicht für den Pöbel gedacht.",
                    )
                    return

                # print(f'Message: {message.text}')
                command_parts = message.text.strip().split()
                if len(command_parts) < 2:
                    self.bot.reply_to(
                        message,
                        "Jo, do muast jetzt scho dazuaschreiben wost willst. Probiers numoi",
                    )
                    return

                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.genericLeistungsmessage,
                    message.chat.id,
                )

                self.bot.reply_to(
                    message,
                    "Wüst des auf irgend an poll replyen?",
                    reply_markup=self.helper.open_polls_button(True),
                )

            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state="*", commands=["cancel"])
        def cancel(message):
            try:
                self.process_cancle(message)
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.normalLocation)
        def get_poll_location(message):
            try:
                location = self.process_poll_location(message)
                if location:
                    self.poller = PersistantLeistungsTagPoller(
                        self.helper,
                        message.chat.id,
                        location,
                        LeistungsTyp.NORMAL,
                    )
                    self.helper.bot.reply_to(
                        message,
                        "Für wann wollen ma pollen?",
                        reply_markup=self.helper.date_suggester(),
                    )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.konkurrenzLocation)
        def get_konkurrenz_location(message):
            try:
                location = self.process_poll_location(message)
                if location:
                    self.poller = PersistantLeistungsTagPoller(
                        self.helper,
                        message.chat.id,
                        location,
                        LeistungsTyp.KONKURENZ,
                    )
                    self.helper.bot.reply_to(
                        message,
                        "Für wann wollen ma pollen?",
                        reply_markup=self.helper.date_suggester(),
                    )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.zusatzLocation)
        def get_zusatz_location(message):
            try:
                location = self.process_poll_location(message)
                if location:
                    self.poller = PersistantLeistungsTagPoller(
                        self.helper,
                        message.chat.id,
                        location,
                        LeistungsTyp.ZUSATZ,
                    )
                    self.helper.pick_date(message.chat.id)
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.removeLocation)
        def remove_location(message):
            try:
                self.helper.remove_location(message.text)
                bot.reply_to(message, "Hab de location murz destroyed!")
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error plox\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.searchLocation)
        def search_location(message):
            try:
                self.process_search_location(
                    message.chat.id,
                    message.text.strip(),
                )
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error (LeistungsState.searchLocation)\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.rateLocation)
        def rate_location(message):
            try:
                rating = self.helper.get_rating(message.text)
                leistungstag = self.helper.db.getLeistungsTags(
                    LeistungsTyp.NORMAL,
                    max_results=1,
                    before=datetime.now(),
                )[0]
                self.helper.db.addUser(message.from_user.id, message.chat.id)
                try:
                    self.helper.db.rateLocationKey(
                        leistungstag["location"],
                        message.from_user.id,
                        rating,
                    )
                except BaseException:
                    self.bot.send_message(
                        message.chat.id,
                        "WAHLBETRUG!! Du host schomoi obgstimmt.",
                    )
                self.bot.delete_state(message.from_user.id)
            except Exception as error:
                bot.send_message(
                    lc.config["chat_id"],
                    f"Hi Devs!!\nHandle This Error (LeistungsState.searchLocation)\n{error}",
                )
                bot.reply_to(message, f"An error occurred!\nError: {error}")
                bot.send_message(
                    lc.config["chat_id"],
                    f"An error occurred!\nError: {error}",
                )

        @bot.message_handler(state=LeistungsState.switcherooLeistungstagNumber)
        def switcheroo_leistungstag_number(
            message: telebot.types.Message,
        ) -> None:
            try:
                lt_number = int(message.text)
            except BaseException:
                self.bot.send_message(
                    message.chat.id,
                    "Host du in da Voikschui ned aufpasst wos a nummer is? Probiers numoi ...",
                )

            lt = self.helper.db.getLeistungstagByNumber(lt_number)
            if lt is None:
                self.bot.send_message(
                    message.chat.id,
                    "Den Leistungstog find i ned. Schau numoi genau",
                )
                return

            print(f'Location {lt["location"]}')

            if message.from_user.id not in self.user_context:
                self.user_context[message.from_user.id] = {"leistungstag": lt}
            else:
                self.user_context[message.from_user.id]["leistungstag"] = lt

            self.bot.send_message(
                message.chat.id,
                "Passt. Wo schau ma stottdessen hin?",
                reply_markup=self.helper.location_keyboard(),
            )
            self.bot.set_state(
                message.from_user.id,
                LeistungsState.switcherooAlternateLocation,
                message.chat.id,
            )

        @bot.message_handler(state=LeistungsState.switcherooAlternateLocation)
        def switcheroo_alternate_location(
            message: telebot.types.Message,
        ) -> None:
            if (
                message.from_user.id not in self.user_context
                or "leistungstag"
                not in self.user_context[message.from_user.id]
                or self.user_context[message.from_user.id]["leistungstag"]
                is None
            ):
                self.bot.send_message(
                    message.chat.id,
                    "Could not find Leistungstag in UserContext. This should not happen, please try again ...",
                )
                self.bot.delete_state(message.from_user.id, message.chat.id)
                return

            location = message.text.strip()
            # check if location exists in database
            info = self.helper.db.getLocationInfo(location)

            if not info:
                self.bot.send_message(
                    message.chat.id,
                    f"'{location}' kenn i ned..wüstas stattdessn zur listn dazua gebn?",
                    reply_markup=self.helper.unkown_location_button(location),
                )
                self.bot.set_state(
                    message.from_user.id,
                    LeistungsState.searchLocation,
                    message.chat.id,
                )

            else:
                lt = self.user_context[message.from_user.id]["leistungstag"]
                self.helper.db.switchLeistungstagLocation(
                    lt["key"],
                    lt["location"],
                    info["key"],
                )

                self.bot.send_message(
                    message.from_user.id,
                    f"Ok, donn gemma am {lt['date'].strftime('%d.%m.%Y')} ins {info['name']}",
                )
                self.bot.delete_state(message.from_user.id, message.chat.id)

            self.user_context[message.from_user.id]["leistungstag"] = None
            # TODO: Edit poll message, if possible

        @bot.message_handler(commands="switcheroo")
        def switcheroo(message: telebot.types.Message) -> None:
            if not self.helper.sender_has_permission(message):
                self.bot.reply_to(
                    message,
                    "Diese Funktion ist nicht für den Pöbel gedacht.",
                )

            self.bot.send_message(
                message.chat.id,
                "Wechan muastn ändern? Schick ma de nummer und i schau wos i doan konn.",
            )
            self.bot.set_state(
                message.from_user.id,
                LeistungsState.switcherooLeistungstagNumber,
                message.chat.id,
            )

        # @bot.message_handler(content_types=["text"])
        # def new_msg(message):
        #     try:
        #         if "nude" in message.text:
        #             if message.chat.type != "private":
        #                 if (datetime.now() - self.last_text_nudes) > timedelta(
        #                     days=1,
        #                 ):
        #                     self.last_text_nudes = datetime.now()
        #                     self.process_send_nudes(message.chat.id)
        #             else:
        #                 self.process_send_nudes(message.chat.id)

        #     except Exception as error:
        #         bot.send_message(
        #             lc.config["chat_id"],
        #             f"Hi Devs!!\nHandle This Error (text)\n{error}",
        #         )
        #         bot.reply_to(message, f"An error occurred!\nError: {error}")

        @bot.message_handler(commands=["version"])
        def version(message):
            bot.reply_to(
                message,
                f"LeistungsBot - {_version.__version__}",
            )

    def process_cancle(self, message):
        self.bot.send_message(
            message.chat.id,
            "Halt Stop.",
            reply_markup=telebot.types.ReplyKeyboardRemove(),
        )
        self.bot.delete_state(message.from_user.id, message.chat.id)

    def process_poll_location(self, message):
        location = message.text.strip()
        # check if location exists in database
        info = self.helper.db.getLocationInfo(location)
        if not info:
            self.bot.send_message(
                message.chat.id,
                f"'{location}' kenn i ned..wüstas stattdessn zur listn dazua gebn?",
                reply_markup=self.helper.unkown_location_button(location),
            )
            self.bot.set_state(
                message.from_user.id,
                LeistungsState.searchLocation,
                message.chat.id,
            )
        elif info["visited"]:
            self.bot.reply_to(
                message,
                "Do woan ma schomoi, i hoff du wast wost duast.",
            )
            self.bot.delete_state(message.from_user.id, message.chat.id)
            return location
        else:
            self.bot.delete_state(message.from_user.id, message.chat.id)
            return location
        return None

    def process_reminder(self, message, leistungstag_key):
        if not self.helper.sender_has_permission(message):
            self.bot.reply_to(
                message,
                "Diese Funktion ist nicht für den Pöbel gedacht.",
            )
            return

        leistungstag = self.helper.db.getLeistungstag(leistungstag_key)
        self.bot.send_message(
            lc.config["leistungschat_id"],
            "Reminder. Morgen wird reserviert. Letzte Chance zum Abstimmen 🗳️",
            reply_to_message_id=leistungstag["poll_id"],
        )
        self.bot.send_message(message.chat.id, "Da Reminder is draußen!")

    def process_closepoll(
        self,
        message: telebot.types.Message,
        leistungstag_key: int,
        sneaky: bool = False,
    ) -> None:
        # Check does not work when in callback
        # if not self.helper.sender_has_permission(message):
        #     self.bot.reply_to(
        #         message,
        #         "Diese Funktion ist nicht für den Pöbel gedacht.",
        #     )
        #     return

        leistungstag = self.helper.db.getLeistungstag(leistungstag_key)
        self.helper.db.closeLeistungstag(leistungstag_key)
        try:
            self.bot.stop_poll(
                lc.config["leistungschat_id"],
                leistungstag["poll_id"],
            )
        except BaseException:
            pass
        try:
            self.bot.unpin_chat_message(
                lc.config["leistungschat_id"],
                leistungstag["poll_id"],
            )
        except BaseException:
            pass
        if not sneaky:
            self.bot.send_message(
                lc.config["leistungschat_id"],
                "Schluss, aus, vorbei die Wahl is glaufen und für de de abgstimmt haben is a Platzerl reserviert.",
                reply_to_message_id=leistungstag["poll_id"],
            )
            self.bot.send_message(
                message.chat.id,
                "De Poll is zua. I hoff für dich d Reservierung is scho erledigt!",
            )
        else:
            try:
                sneakiely = (
                    importlib.resources.as_file(
                        importlib.resources.files("resources") / "sneaky.gif",
                    ),
                )
            except BaseException:
                resources = Path(__file__).parent / "resources"
                sneakiely = resources / "sneaky.gif"
            self.bot.send_animation(
                message.chat.id,
                animation=telebot.types.InputFile(
                    sneakiely,
                ),
                caption="De Poll is zua. I hoff für dich d Reservierung is scho erledigt!",
            )

    def process_check_open_hours(self, callback, open_hours_correct):
        if open_hours_correct:
            self.poller.dry_send()

    def process_send_nudes(self, chat_id):
        self.helper.send_nude(chat_id)

    def process_search_location(self, chat_id, query):
        finds, rand_id = self.helper.search_location(query)
        if finds < 1:
            self.bot.send_message(
                chat_id,
                f'Wenn i nach "{query}" suach find i nix...vielleicht verschriebn?',
            )
        elif finds == 1:
            self.helper.approve_location(chat_id, rand_id, 0)
        elif finds > 1:
            self.bot.send_message(
                chat_id,
                "Suach da aus wost willst, oda schick ma wos aunders",
                reply_markup=self.helper.search_location_button(rand_id),
            )

    def process_history(self, message):
        self.bot.send_message(
            message.chat.id,
            "Welche Art von Leistungstag willst da anschaun?",
            reply_markup=self.helper.leistungstag_history_type_button(),
        )

    def process_purge(self, message):
        try:
            responisibility = (
                importlib.resources.as_file(
                    importlib.resources.files("resources")
                    / "responisibility.gif",
                ),
            )
        except BaseException:
            resources = Path(__file__).parent / "resources"
            responisibility = resources / "responisibility.gif"
        self.bot.send_animation(
            message.chat.id,
            telebot.types.InputFile(
                responisibility,
            ),
            caption="Welche Art von Leistungstag willst löschen?",
            reply_markup=self.helper.leistungstag_purge_type_button(),
        )

    def process_generic_leistungsmessage(
        self,
        original_message: telebot.types.Message,
        leistungstag_id,
    ):
        command_parts = original_message.text.strip().split()
        msg = " ".join(command_parts[1:])

        if leistungstag_id is not None:
            leistungstag = self.helper.db.getLeistungstag(leistungstag_id)
            self.bot.send_message(
                lc.config["leistungschat_id"],
                msg,
                reply_to_message_id=leistungstag["poll_id"],
            )
        else:
            self.bot.send_message(lc.config["leistungschat_id"], msg)

        print(
            f'User {original_message.from_user.username} sent generic message "{msg}"',
        )

        self.bot.delete_state(
            original_message.from_user.id,
            original_message.chat.id,
        )

    def check_open_hours_before_sending(self, call, date: date):
        open_state = self.helper.check_open_hours(self.poller.location, date)

        if open_state[0] == Openness.OPEN:
            self.poller.dry_send_with_date(date)
        else:
            self.poller.date = date

            if open_state[0] == Openness.CLOSED:
                self.bot.send_message(
                    call.message.chat.id,
                    "I glaub ned, dass de offn hom. Bist da sicha?\n\n"
                    + open_state[1],
                    reply_markup=self.helper.check_open_hours_keyboard(
                        "Des passt so, i kenn mi aus",
                    ),
                )
            elif open_state[0] == Openness.SHORT:
                self.bot.send_message(
                    call.message.chat.id,
                    "Is da des long gmua?\n\n" + open_state[1],
                    reply_markup=self.helper.check_open_hours_keyboard(
                        "Jo, passt scho",
                    ),
                )
            elif open_state[0] == Openness.UNKNOWN:
                self.bot.send_message(
                    call.message.chat.id,
                    "I was jetzt hod ned, ob de offen hom. Muast söwa schaun.",
                    reply_markup=self.helper.check_open_hours_keyboard(
                        "Des passt so, i kenn mi aus",
                        "Schaut schlecht aus",
                    ),
                )

    def try_remind_to_leistungstag_on_a_specific_date(
        self,
        message: telebot.types.Message,
        target_date: date,
    ) -> bool:
        """! Reminds to polls of leistungstag on a specific date

        @param message Reminder command message
        @param target_date Date of the Leistungstag(e)

        @returns If Leistungstage on this date where found
        """

        if target_date < datetime.now().date():
            self.bot.send_message(
                message.chat.id,
                "Für des is zu spät zum reminden. Suach da wos ondares aus ...",
            )
            return False

        leistungstage = self.helper.db.getLeistungstageByDate(target_date)
        # print(f'Leistungstage: {leistungstage}')

        if leistungstage is None:
            self.bot.send_message(
                message.chat.id,
                "An dem Tog is ka Leistungstog. Suach da wos ondares aus ...",
            )
            return False

        if len(leistungstage) == 1:
            leistunstag = leistungstage[0]
            location_name = self.helper.db.getLocationName(
                leistunstag["location"],
            )

            if leistunstag["closed"] == 1:
                self.bot.reply_to(
                    message,
                    f"Leistungstag an dem Tog is bei {location_name}, aber da Poll is scho geclosed. Willst wirklich on den reminden?",
                    reply_markup=self.helper.confirm_leistungstag_button(
                        leistunstag["key"],
                    ),
                )
            else:
                self.bot.reply_to(
                    message,
                    f"Leistungstag an dem Tog is bei {location_name}. Willst on den reminden?",
                    reply_markup=self.helper.confirm_leistungstag_button(
                        leistunstag["key"],
                    ),
                )

        else:
            self.bot.send_message(
                message.chat.id,
                "Ein wöd Tog. Do san sogor mehrere Leistungstoge!",
            )
            self.bot.reply_to(
                message,
                "An welchen Poll wüst reminden?",
                reply_markup=self.helper.polls_button(leistungstage),
            )

        return True

    def infinite_poll(self):
        self.bot.infinity_polling()

    def poll(self):
        self.bot.polling()


def main():
    parser = argparse.ArgumentParser(
        description="LeistungsBot - some realy weird bot.",
    )
    parser.add_argument(
        "--google",
        "-g",
        dest="google",
        help="Google api key",
    )
    parser.add_argument(
        "--token",
        "-t",
        dest="bot_token",
        help="Telegram bot token generated by @BotFather",
    )
    parser.add_argument(
        "--hash",
        dest="api_hash",
        help="api hash from my.telegram.org",
    )
    parser.add_argument(
        "--id",
        dest="api_id",
        help="api id from my.telegram.org",
    )
    parser.add_argument(
        "--chat",
        dest="chat_id",
        help="chat id for your private group, to view logs and errors",
    )
    parser.add_argument(
        "--leistungchat",
        dest="leistungschat_id",
        help="chat id from the main group",
    )
    parser.add_argument(
        "--eistungsadmin",
        dest="leistungsadmin_id",
        help="chat id from the admin group",
    )
    parser.add_argument(
        "--host",
        dest="mysql.host",
        help="host name/ip from the mysql server",
    )
    parser.add_argument(
        "--db",
        dest="mysql.db",
        help="databse name",
    )
    parser.add_argument(
        "--user",
        dest="mysql.user",
        help="user with access to the databse",
    )
    parser.add_argument(
        "--password",
        dest="mysql.password",
        help="user password",
    )
    parser.add_argument("--config", "-c", help="Provide a custom config file")

    args = parser.parse_args()
    lc.set_args(args, dots=True)
    print("Starting LeistungsBot")
    lb = LeistungsBot()
    lb.infinite_poll()
